#include "TransformResiduals.h"
#include "MathSVD.h"
#include "OutputKin.h"
#include "PreMeta.h"

bool FastTransform::empKin = false;
bool FastTransform::pedKin = false;
String FastTransform::readInEmp = "";

   FastTransform::FastTransform()
:pPheno			(NULL)
{}

FastTransform::~FastTransform()
{ 
   if(pPheno) delete [] pPheno;
}

void FastTransform::GetFixedEffectsCount(Pedigree & ped, bool useCovariates)
{
   fixedEffects = 1 + (useCovariates ? ped.covariateCount : 0);
}

//Prepare is to get pPheno filled up and useful persons and families clarified.
void FastTransform::Prepare(Pedigree & ped, int traitNum,bool useCovariates)
{
   pPheno = new IntArray [ped.familyCount];

   //if genotpye is read from vcf file, need to make sure to eliminate individuals who are not sequenced.

   for(int p=0;p<ped.count;p++)
   {
      samplePEDIDHash.SetInteger(ped[p].pid,p);
      if(ped[p].isFounder())
	 foundersHash.SetInteger(ped[p].pid,p);
   }

   if(PreMeta::genoFromVCF)
   {
      VcfFileReader reader;
      VcfHeader header;
      reader.open(PreMeta::vcfInput,header);
      int numSamples = header.getNumSamples();

      for(int s=0;s<numSamples;s++)
      {
	 const char * sample = header.getSampleName(s);
	 sampleVCFIDHash.SetInteger(sample,s);
      }

      for (int f = 0; f < ped.familyCount; f++)
      {
	 pPheno[f].Dimension(0);
	 for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++) 
	 {
	    if (ped[i].isPhenotyped(traitNum) && (!useCovariates || ped[i].isFullyControlled()))
	    {
	       int p = sampleVCFIDHash.Integer(ped[i].pid);
	       if(p!=-1)
		  pPheno[f].Push(i);
	    }
	 }
      }
      reader.close();
   }
   else
   {
      // If we are using covariates, we only consider individuals
      // for which all covariates have been recorded
      for (int f = 0; f < ped.familyCount; f++)
      {
	 pPheno[f].Dimension(0);
	 for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++) {
	    if (ped[i].isPhenotyped(traitNum) && (!useCovariates || ped[i].isFullyControlled()))
	    {
	       pPheno[f].Push(i);
	       //printf("trait of person %s is: %g\n",ped[i].pid,ped[i].traits[traitNum]);
	    }
	 }
      }
   }
   // Number of families with non-null kinships and phenotypes
   families = persons = 0;
   // Count useful families
   for (int f = 0; f < ped.familyCount; f++)
   {
      if (pPheno[f].Length()) {
	 families++;
	 persons += pPheno[f].Length();
      }
   }
   if (families == 0)
   {
      printf("Trait: %s -- No informative families\n\n",
	    (const char *) ped.traitNames[traitNum]);
   }
   printf("# of useful families for trait %s is: %d; # of useful persons is: %d.\n",ped.traitNames[traitNum].c_str(),families,persons);

   Vector trait; 
   trait.Dimension(persons); 
   int index=0;
   for (int f = 0; f < ped.familyCount; f++)
   {	    
      for(int j=0;j<pPheno[f].Length();j++) 
      {
	 if(ped[pPheno[f][j]].isFounder())
	    founders.Push(ped[pPheno[f][j]].pid);
	 trait[index]=ped[pPheno[f][j]].traits[traitNum]; 
	 index++;
      }
   }
   traitVar = trait.Var();
   traitMean=trait.Average();
   UY.Dimension(persons);
   UY.Zero();
   UX.Dimension(persons,fixedEffects);
   Y.Dimension(persons);
   UDY.Dimension(persons);
   UDX.Dimension(persons,fixedEffects);
   UD.Dimension(persons,persons);
   inv.Dimension(fixedEffects,fixedEffects);

   if(OutputKin::outputKin || OutputKin::outputKinX) 
   {
      //OutputKin outkin;
      //outkin.Output(ped);
   }
   //Fill in X and Y
   // Phenotypes
   index=0;
   for (int f = 0; f < ped.familyCount; f++) {
      int ct = pPheno[f].Length();
      for(int i=0;i<ct;i++)
      {
	 Y[index]=ped[pPheno[f][i]].traits[traitNum];
	 index++;
      }
   }
   // Setup matrix of fixed effects for each family
   X.Dimension(persons, fixedEffects);
   for (int i = 0; i < persons; i++)
   {
      // Constant for regressing grand mean
      X[i][0] = 1.0;
   }

   if (useCovariates)
   {
      int index =0;
      for(int f=0;f<ped.familyCount;f++){
	 int count = pPheno[f].Length();
	 for(int i=0;i<count;i++){
	    // User specified covariates
	    for (int j = 1; j <= ped.covariateCount; j++)
	       X[index][j] = ped[pPheno[f][i]].covariates[j-1];
	    index++;
	 }
      }
   }
}

void FastTransform::TransformPedkinSibEnv(AutoFit & engine, Pedigree &ped)
{
   //Setup Omega
   Kinship kin;
   KinshipX kinX;
   Matrix omega;
   SVD svd;
   for (int f = 0; f < ped.familyCount; f++)
   {
      if (pPheno[f].Length() > 0)
      {
	 int count = pPheno[f].Length();
	 kin.Setup(*ped.families[f]);
	 if(AutoFit::fitX)
	    kinX.Setup(*ped.families[f]);
	 omega.Dimension(count,count);

	 for (int i = 0; i < count; i++)
	 {
	    //Get the diagnol elements
	    if(AutoFit::fitX)
	    {
	       omega[i][i] = 
		  (2.0*kin(ped[pPheno[f][i]], ped[pPheno[f][i]])*engine.heritability
		   + 2.0*kinX(ped[pPheno[f][i]], ped[pPheno[f][i]])*engine.sigma_gX + engine.sharedEnv + engine.sigma_e2) * traitVar;
	    }
	    else
	    {
	       omega[i][i] = 
		  (2.0*kin(ped[pPheno[f][i]], ped[pPheno[f][i]])*engine.heritability
		   + engine.sharedEnv + engine.sigma_e2) * traitVar;
	    }

	    //Get the off-diagnoal elements
	    if(ped[pPheno[f][i]].zygosity==999)
	    {
	       for (int j = 0; j < count; j++)
	       {
		  if(i!=j)
		  {
		     if(AutoFit::fitX)
		     {
			omega[i][j] = omega[j][i] =
			   ped[pPheno[f][i]].isSib(ped[pPheno[f][j]])*engine.sharedEnv * traitVar;
		     }
		     else
		     {
			omega[i][j] = omega[j][i] =
			   + ped[pPheno[f][i]].isSib(ped[pPheno[f][j]])*engine.sharedEnv*traitVar;
		     }
		  }
	       }
	    }
	    else
	    {
	       for(int j=i+1;j<count;j++)
	       {
		  if(AutoFit::fitX)
		  {
		     omega[i][j] = omega[j][i] =
			(2.0*kin(ped[pPheno[f][i]], ped[pPheno[f][j]])*engine.heritability
			 + 2.0*kinX(ped[pPheno[f][i]], ped[pPheno[f][j]])*engine.sigma_gX + ped[pPheno[f][i]].isSib(ped[pPheno[f][j]])*engine.sharedEnv) * traitVar;
		  }
		  else
		  {
		     omega[i][j] = omega[j][i] =
			(2.0*kin(ped[pPheno[f][i]], ped[pPheno[f][j]])*engine.heritability
			 + ped[pPheno[f][i]].isSib(ped[pPheno[f][j]])*engine.sharedEnv)*traitVar;
		  }
	       }
	    }
	 }

	 svd.Decompose(omega);
	 if(transU.rows==0)
	 {
	    transU.Transpose(svd.u);
	    D.Copy(svd.w);
	 }
	 else
	 {
	    int oldDim = transU.rows;
	    transU.GrowTo(oldDim,oldDim+count,0.0);
	    transU.GrowTo(oldDim+count,oldDim+count,0.0);
	    D.GrowTo(transU.cols);
	    for(int i=oldDim;i<oldDim+count;i++)
	    {
	       D[i] = svd.w[i-oldDim]; 
	       for(int j=oldDim;j<transU.cols;j++)
		  transU[i][j] = svd.u[j-oldDim][i-oldDim]; 
	    }
	 }
      }
   }
   //start doing transformation
   UX.Product(transU,X);
   for(int r=0;r<transU.rows;r++)
      UY[r] = transU[r].InnerProduct(Y);
}

void FastTransform::TransformPedkin(Pedigree & ped,int traitNum,bool useCovariates)
{
   printf("Now transforming pedigree kinship matrix ...\n");
   transU.Dimension(persons,persons);
   transU.Zero();
   D.Dimension(persons);
   int index=0;
   for (int f = 0; f < ped.familyCount; f++)
   {
      if (pPheno[f].Length() > 0)
      {
	 int count = pPheno[f].Length();
	 //printf("fam %d has %d members.\n",f,count);
	 Vector tmppheno;
	 tmppheno.Dimension(count);
	 for (int i = 0; i < count; i++)
	 {
	    tmppheno[i] = ped[pPheno[f][i]].traits[traitNum];
	 }

	 // Setup matrix of fixed effects for each family
	 Matrix sub_X;
	 sub_X.Dimension(count, fixedEffects);
	 Matrix transUX;

	 //fill in sub_X
	 for (int i = 0; i < count; i++)
	 {
	    // Constant for regressing grand mean
	    sub_X[i][0] = 1.0;

	    // User specified covariates
	    if (useCovariates)
	       for (int j = 1; j <= ped.covariateCount; j++)
		  sub_X[i][j] = ped[pPheno[f][i]].covariates[j-1];
	 }

	 //fill in X
	 if(X.rows==0)
	 {
	    X.Copy(sub_X);
	 }
	 else
	 {
	    X.StackBottom(sub_X);
	 }

	 Kinship kin;
	 kin.Setup(*ped.families[f]);

	 //Output kinship matrix

	 Matrix pedkin;
	 pedkin.Dimension(count,count);
	 for (int i = 0; i < count; i++)
	    for (int j = i; j < count; j++)
	       pedkin[i][j] = pedkin[j][i] =
		  2.0*kin(ped[pPheno[f][i]], ped[pPheno[f][j]]);

	 //refine kinship matrix for adopted children
	 for (int i = 0; i < count; i++)
	 {
	    if(ped[pPheno[f][i]].zygosity==999)
	    {
	       for(int j=0;j<count;j++)
	       {
		  if(i!=j)
		  {
		     pedkin[i][j] = pedkin[j][i] = 0.0;
		  }
	       }
	    }
	 }
	 /*
	    for(int i=0;i<count;i++)
	    {
	    for(int j=0;j<count;j++)
	    printf("%g ",pedkin[i][j]);
	    printf("\n");
	    }
	  */
	 //now do the transformation
	 //printf("Now doing the cholesky decompositon ... /n");
	 //S is the vector of eigenvalues
	 SVD svd;
	 svd.Decompose(pedkin);
	 Vector S;
	 S.Dimension(count);
	 Matrix U,transUsub;
	 U.Copy(svd.u);
	 S.Copy(svd.w);
	 transUsub.Transpose(U);
	 transUX.Product(transUsub,sub_X);

	 //fill up UX, D, and Y
	 for(int row=0;row<count;row++)
	 {
	    D[row+index] = S[row];
	    Y[row+index] = tmppheno[row];
	    for(int col=0;col<fixedEffects;col++)
	    {
	       UX[row+index][col] = transUX[row][col];
	    }
	 }

	 int block = index;
	 //calculate transpose(U)*y and save them in the traits vector
	 for(int r=0;r<count;r++)
	 {	
	    //fill up transU here
	    for(int c =0;c<count;c++)
	       transU[r+block][c+block] = transUsub[r][c];
	    //fill up UY here
	    for(int c=0;c<count;c++)
	       UY[index] += transUsub[r][c]*tmppheno[c];
	    index++;
	    //if(index>=2477 && index<=2480 || index>=2490 && index<=2492)
	    //    printf("%d pid is: %s\n",index,ped[pPheno[f][r]].pid.c_str());
	 }
      }
   }
} 

void FastTransform::TransformEmpkin(Pedigree & ped,int traitNum, bool useCovariates)
{
   transU.Dimension(persons,persons);
   D.Dimension(persons);
   SVD svd;

   //if reading in Empirical matrix from outside
   if(readInEmp != "")
   {
      FILE * empIn;

      IntArray selectedSample;
      StringIntHash IDFromEmp;
      StringArray tmp;
      //Get the pids from kinship matrix
     // printf("Reading in sample IDs and get empirical kinship setup ...\n");
      empIn = fopen(readInEmp,"r");
      int idx=0;
      while (!feof(empIn))
      {
	 if(idx==0)
	 {
	    String buffer;
	    buffer.ReadLine(empIn);
	    tmp.Clear();
	    tmp.AddTokens(buffer, SEPARATORS);
	    for(int i=0;i<tmp.Length();i++)
	       IDFromEmp.SetInteger(tmp[i],i);
	    idx++;
	 }
	 else 
	    break;
      }
      fclose(empIn);

      Matrix allPairs;
      //the dimension of allPairs should be consistent to what is in the file
      allPairs.Dimension(tmp.Length(),tmp.Length());

      //Check if all individuals from pPheno are included in read-in Kinship.
      printf("Checking if all individuals are included in read-in kinship ...\n");
      for (int f = 0; f < ped.familyCount; f++)
      {
	 for(int j=0;j<pPheno[f].Length();j++)
	 {
	    int p = IDFromEmp.Integer(ped[pPheno[f][j]].pid);
	    if(p==-1)
	       error("Fatal Error: person %s in ped file is not included in read-in kinship. The program will stop here. Please fix the problem and re-run.\n",ped[pPheno[f][j]].pid.c_str());
	    else 
	    {
	       selectedSample.Push(p);
	    }
	 }
      }

      //printf("Now reading in empirical kinship ...\n");
      FILE * newEmpIn;
      newEmpIn = fopen(readInEmp,"r");
      idx=0;
      while (!feof(newEmpIn))
      {
	 idx++;
	 String buffer;
	 buffer.ReadLine(newEmpIn);
         if(idx==1)
            continue;
	 tmp.Clear();
	 tmp.AddTokens(buffer, SEPARATORS);
	 for(int i=0; i<tmp.Length(); i++)
	 {
	    allPairs[idx-2][i] = tmp[i].AsDouble();
	 }
      }
      fclose(newEmpIn);  

      //printf("Subsetting ...\n");
      Matrix sub_allPairs;
      sub_allPairs.Dimension(persons,persons);
      for(int r=0;r<selectedSample.Length();r++)
      {
	 for(int c=0;c<selectedSample.Length();c++)
	 {
	    sub_allPairs[r][c] = allPairs[selectedSample[r]][selectedSample[c]];
	 }
      }
      allPairs.Copy(sub_allPairs);

      printf("Now transforming empirical matrix ...\n");
      svd.Decompose(allPairs);
      printf("Transformation of empirical kinship done.\n");
   }
   else
   {
      //printf("Calculating empirical kinship matrix ...\n");
      KinshipEmp bnKin;

      if(PreMeta::genoFromVCF || PreMeta::dosage)
	 bnKin.SetupVCF(ped,foundersHash);
      else
	 bnKin.Setup(ped);

      if(persons< bnKin.allPairs.rows)
	 SubSet(bnKin.allPairs,ped);

      printf("Now transforming empirical matrix ...\n");
      svd.Decompose(bnKin.allPairs);
      printf("Transformation of empirical kinship done.\n");
   }

   transU.Transpose(svd.u);
   D.Copy(svd.w);
   UX.Product(transU,X);
   //calculate transpose(U)*y and save them in the traits vector
   for(int r=0;r<transU.rows;r++)
      UY[r] = transU[r].InnerProduct(Y);
}

void FastTransform::SubSet(Matrix & allPairs,Pedigree & ped)
{
   IntArray sampleOrdered;
   if(PreMeta::genoFromPed)
   {
      for (int f = 0; f < ped.familyCount; f++)
      {
	 for(int i=0;i<pPheno[f].Length();i++)
	 {
	    sampleOrdered.Push(pPheno[f][i]);
	 }
      }
   }

   if(PreMeta::genoFromVCF || PreMeta::dosage)
   {
      for (int f = 0; f < ped.familyCount; f++)
      {
	 for(int j=0;j<pPheno[f].Length();j++)
	 {
	    //get sample index from vcf using hash
	    int p = sampleVCFIDHash.Integer(ped[pPheno[f][j]].pid);
	    if(p!=-1)
	    {
	       sampleOrdered.Push(p);
	    }
	 }
      }
   }

   Matrix tmp;
   tmp.Dimension(persons,persons);
   //printf("sampleOrdered length: %d; persons: %d\n",sampleOrdered.Length(),persons);
   for(int r=0;r<sampleOrdered.Length();r++)
   {
      for(int c=0;c<sampleOrdered.Length();c++)
      {
	 tmp[r][c] = allPairs[sampleOrdered[r]][sampleOrdered[c]];
      }
   }
   allPairs.Copy(tmp);
}

//This is done after VCs have been fit
void FastTransform::FinalizeTransform(double sigma_g2, double sigma_e2)
{
   if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
   {
      for(int i=0;i<persons;i++)
      {
	 for(int j=0;j<persons;j++)
	 {
	    UD[i][j]=transU[i][j]/sqrt(D[i]);
	 }
      }
   }
   else
   {
      for(int i=0;i<persons;i++)
      {
	 for(int j=0;j<persons;j++)
	 {
	    UD[i][j]=transU[i][j]/sqrt(sigma_g2*D[i]+sigma_e2);
	 }
      }
   }
   /*
   //fill in UDY
   for(int i=0;i<persons;i++)
   {
   UDY[i]=UD[i].InnerProduct(Y);
   }
   //fill in UDX
   UDX.Product(UD,X);
    */
   if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
   {
      for(int i=0;i<persons;i++)
      {
	 for(int j=0;j<fixedEffects;j++)
	 {
	    UDX[i][j]=UX[i][j]/sqrt(D[i]);
	 }
	 UDY[i]=UY[i]/sqrt(D[i]);
      }
   }
   else
   {
      for(int i=0;i<persons;i++)
      {
	 for(int j=0;j<fixedEffects;j++)
	 {
	    UDX[i][j]=UX[i][j]/sqrt(sigma_g2*D[i]+sigma_e2);
	 }
	 UDY[i]=UY[i]/sqrt(sigma_g2*D[i]+sigma_e2);
      }
   }

   Matrix tmp_i;
   tmp_i.Dimension(fixedEffects,fixedEffects);
   inv.Zero();

   for(int i=0;i<persons;i++)
   {
      for(int row=0;row<fixedEffects;row++)
      {
	 for(int col=0;col<fixedEffects;col++)
	 {
	    tmp_i[row][col] = UDX[i][row]*UDX[i][col];
	 }
      }
      inv.Add(tmp_i);
   }

   SVD svd;
   svd.InvertInPlace(inv); //now inv is completed
   //printf("Finished finalizing the transformations...\n");
}

void FastTransform::ReformCholesky(Matrix & M, Vector & D)
{
   if(M.rows != M.cols) 
   {
      printf("Error! Matrix is not square.\n");
   } else {
      for(int i=0;i<M.rows;i++)
      {
	 D[i] = M[i][i] * M[i][i];
	 M[i][i] = 1.0;
      }
   }
}
